/* 
 * Mach Operating System
 * Copyright (c) 1993 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Carnegie Mellon requests users of this software to return to
 * 
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 * 
 * any improvements or extensions that they make and grant Carnegie Mellon
 * the rights to redistribute these changes.
 *
 *
 * This is a hack,  mostly borrowed from the dosfs source.
 *                                    Morgan Price, Nov 93
 */

#include <sys/param.h>
#include <sys/dir.h>
#include <sys/user.h>
#include <sys/buf.h>
#include <sys/inode.h>
#include <sys/systm.h>
#include <sys/kernel.h>

#include "macfs_internals.h"

#include "vnode_manage_doc.h"
#include "vnode_calls_doc.h"
#include "system_calls_doc.h"

struct vfsops macfs_vfsops = {
	macfs_mount,
	macfs_unmount,
	macfs_root,
	macfs_statfs,
	macfs_sync,
	macfs_fhtovp
};

struct vnodeops macfs_vnodeops = {
	macfs_open,		/* open */
	macfs_close,		/* close */
	macfs_rdwr,		/* rdwr */
	macfs_ioctl,		/* ioctl */
	macfs_select,		/* select */
	macfs_getattr,		/* getattr */
	macfs_setattr,		/* setattr */
	macfs_access,		/* access */
	macfs_lookup,		/* lookup */
	macfs_create,		/* create */
	macfs_remove,		/* remove */
	(void *)einval,		/* link */
	macfs_rename,		/* rename */
	macfs_mkdir,		/* mkdir */
	macfs_rmdir,		/* rmdir */
	macfs_readdir,		/* readdir */
	macfs_symlink,		/* symlink */
	macfs_readlink,		/* readlink */
	(void *)einval,		/* syncip */
	macfs_inactive,		/* inactive */
	(void *)einval,		/*macfs_nfsbmap*/
	macfs_strategy,		/* strategy */
	(void *)einval,		/* bread */
	(void *)einval,		/*macfs_brelse*/
	(void *)einval,		/* lockctl */
	macfs_fid,		/* fid */
	(void *)einval,		/*macfs_pagein page read*/
	(void *)einval,		/*macfs_pageout page write*/
	macfs_readdir,		/* read1dir */
	macfs_freefid,		/* freefid */
	macfs_nlinks,		/* nlinks */
	macfs_vupdate,		/* iupdate */
};

int macfs_debug = 1;

#define trace(_l_,_p_)	if (_l_) { printf _p_; printf("\n"); }

#define CallMac(Command, Error, Desc)\
{\
    int mac_error_code;\
    if(macfs_debug)\
    {\
	printf("Calling macfs ");\
	printf Desc;\
	printf("\n");\
    }\
    mac_error_code = Command;\
    if(macfs_debug || mac_error_code != 0)\
    {\
	printf("Macfsreturned ");\
	printf Desc;\
	printf(": %d\n", mac_error_code);\
    }\
    if(mac_error_code != 0) Error;\
}\

/*
 * Flag to allow forcible unmounting.
 */
int macfs_doforce = 1;

/*
 * Called by vfs_mountroot to check if the
 * root filesystem can be mounted by us.
 */
extern struct vnode *bdevvp();

/* Probably doesn't work */
int macfs_mountroot()
{
	register struct vfs *vfsp;
	extern struct vnode *rootvp;
	struct macfs_mount *macfsmp;
	register struct fs *fs;
	int error;
	struct macfs_args args;

trace(macfs_debug,("\nmacfs_mountroot "));

	ZALLOC(vfs_vfs_zone, vfsp, struct vfs *);
	VFS_INIT(vfsp, &macfs_vfsops, (caddr_t)0);

	vfsp->vfs_flag = 0;
	vfsp->vfs_exroot = 0;
	vfsp->vfs_vnodecovered = NULL;

	if (rootdev != NODEV)
		rootvp = bdevvp(rootdev);
	args.flags = 0;
	error = macfs_mountfs(&rootvp, "/", vfsp, &args);
	if (error) {
		ZFREE(vfs_vfs_zone, vfsp);
		return (error);
	}
	error = vfs_add((struct vnode *)0, vfsp, 0);
	if (error) {
		(void)macfs_unmount(vfsp);
		ZFREE(vfs_vfs_zone, vfsp);
		return (error);
	}
	/* vfs_add clobbers this */
	vfsp->vfs_devvp = rootvp;

	macfsmp = VFSTOMACFS(vfsp);

	vfs_unlock(vfsp);
	/* no inittodr(), there is no disk time */
	return (0);
}

/*
 * VFS Operations.
 *
 * mount system call
 */
int macfs_mount(
	struct vfs *vfsp,
	char *path,
	caddr_t data)
{
	struct vnode *devvp;
	struct macfs_args args;
	int error;
	dev_t dev;

	trace(macfs_debug,("\nmacfs_mount on %s\n", path));
	/*
	 * Get arguments
	 */
	if (error = copyin(data, &args, sizeof args)) {
		/* compat */
		bzero(&args,sizeof(args));
		if (error = copyin(data, &args, sizeof(struct ifs_args)))
			return error;
	}

	if ((error = getmdev(args.fspec, &dev)) != 0)
		return (error);
	/*
	 * make a special (device) vnode for the filesystem
	 */
	devvp = bdevvp(dev);
trace(macfs_debug,("devvp: %x %d\n", devvp, devvp->v_count));

	/*
	 * Mount the filesystem.
	 */
	error = macfs_mountfs(&devvp, path, vfsp, &args);
	if (error) {
		VN_RELE(devvp);
	}
	return (error);
}

/*
 * Common code for mount and mountroot
 * Actually macfs dependent
 */
int macfs_mountfs(
	struct vnode **devvpp,
	char		*path,
	struct vfs	*vfsp,
	struct macfs_args *args)
{
    register struct vnode *devvp = *devvpp;
    register struct macfs_mount *macfsmp = (struct macfs_mount *)0;
    dev_t dev = devvp->v_rdev;
    int error = EINVAL, i;
    int needclose = 0;
    extern struct vnode *rootvp;
    int bits = VREAD | (vfsp->vfs_flag & VFS_RDONLY? 0 : VWRITE);

    /*
     * Disallow multiple mounts of the same device.
     * Disallow mounting of a device that is currently in use
     * (except for root, which might share swap device for miniroot).
     * Flush out any old buffers remaining from a previous use.
     */
    if (devvp->v_vfsmountedhere && devvp->v_vfsmountedhere != vfsp)
	return (EBUSY);
    if (devvp->v_count > 1 && devvp != rootvp)
	return (EBUSY);
    binval(devvp);
    if (error = VOP_OPEN(devvpp, bits, u.u_cred))
	return (error);
    needclose = 1;

    macfsmp = (struct macfs_mount *)malloc(sizeof *macfsmp);
    macfsmp->vfsp = vfsp;
    if (args->flags & MACFSMNT_PROT) {
	macfsmp->im_uid = args->uid;
	macfsmp->im_gid = args->gid;
	macfsmp->im_mode = args->mode;
    } else {
	macfsmp->im_uid = u.u_uid;
	macfsmp->im_gid = u.u_gid;
	macfsmp->im_mode = bits | (bits >> 3) | (bits >> 6);
    }

    /* Now we need to make the mount information */
    CallMac(macmount(devvp, args->partition, &macfsmp->mrec),
	    error = EIO,
	    ("Calling macmount on partition %d, mrec %x",
	     args->partition, &macfsmp->mrec));
    if(error)
	goto out;
    return (0);

out:
    if (needclose)
	(void)VOP_CLOSE(devvp, FREAD, u.u_cred);
    if (macfsmp && error) {
	free((caddr_t)macfsmp);
	vfsp->vfs_data = (caddr_t)0;
    }
    return (error);
}

/*
 * unmount system call
 */
int macfs_unmount(
	struct vfs	*vfsp)
{
	register struct macfs_mount *macfsmp;
	struct vnode *devvp;
	int i, error = 0;

trace(macfs_debug,("\nmacfs_unmount "));

	macfsmp = VFSTOMACFS(vfsp);

	/* flush everything, check that we can exit */
	CallMac(syncmountrecord(&macfsmp->mrec), ;,
		("syncmountrecord %x", &macfsmp->mrec));
	i = macfs_iflush(macfsmp->im_dev);
	if (i < 0 && !macfs_doforce)
		return (EBUSY);

	if (i == 0) {
		struct vnode *devvp = macfsmp->im_devvp;

		error = VOP_CLOSE(devvp,
			(vfsp->vfs_flag & VFS_RDONLY) ? FREAD : FREAD|FWRITE,
			u.u_cred);
		tbflush(devvp, 0);
		binval(devvp);
		binvalfree(devvp);
		VN_RELE(devvp);
		CallMac(macunmount(&macfsmp->mrec), ;,
			("macunmount %x", &macfsmp->mrec));
		free((caddr_t)macfsmp);
	}
	return (error);
}

int macfs_root(
	struct vfs *vfsp,
	struct vnode **vpp)
{
    struct macfs_node *mp;
    struct macfs_mount *macfsmp = VFSTOMACFS (vfsp);

trace(macfs_debug,("\nmacfs_root "));
    *vpp = MAC_TOV(macfs_finddir(macfsmp, ROOTID));
    return(*vpp? 0 : ENODEV);
}

/*
 * Get file system statistics.
 */
int macfs_statfs(
	struct vfs	*vfsp,
	register struct statfs *sbp)
{
	register struct macfs_mount *macfsmp;
	register struct fs *fs;

trace(macfs_debug,("\nmacfs_statfs "));
	macfsmp = VFSTOMACFS(vfsp);

	sbp->f_type = MOUNT_MACFS;
	sbp->f_bsize = macfsmp->mrec.vib.AlBlkSiz;
	sbp->f_blocks = macfsmp->mrec.vib.NmAlBlks;
	sbp->f_bfree = macfsmp->mrec.vib.FreeBks;
	sbp->f_bavail = macfsmp->mrec.vib.FreeBks;
	sbp->f_files = macfsmp->mrec.vib.FilCnt;
	sbp->f_ffree = macfsmp->mrec.vib.FilCnt;

	/* XXX This should be a real fsid, but this will do for now */
	bcopy((caddr_t)&vfsp->vfs_fsid,
	      (caddr_t)&sbp->f_fsid, sizeof (fsid_t));

	return (0);
}

/*
 * Save superblock info on disk
 */
int macfs_sync(
	struct vfs *vfsp)
{
    CallMac(syncmountrecord(&VFSTOMACFS(vfsp)->mrec), return(EIO),
	    ("syncmountrecord %x", &VFSTOMACFS(vfsp)->mrec));
    return(0);
}

int macfs_open(
	struct vnode	**vpp,
	int		flag,
	struct ucred	*cred)
{
trace(macfs_debug,("\nDopen "));
        VN_HOLD(*vpp);
	return (0);
}

/*
 * Close called
 *
 * Update the times on the inode on writeable file systems.
 */
int macfs_close(
	struct vnode	*vp,
	int		fflag,
	struct ucred	*cred)
{
trace(macfs_debug,("\nDclose "));
        VN_RELE(vp);
	return (0);
}

/*
 * Check mode permission on inode pointer. Mode is READ, WRITE or EXEC.
 * The mode is shifted to select the owner/group/other fields. The
 * super user is granted all permissions.
 */
int macfs_access(
	struct vnode	*vp,
	register int	mode,
	struct ucred	*cred)
{
	register struct macfs_node *ip;
	int error;

trace(macfs_debug,("Daccess "));

	ip = VTO_MAC(vp);
	macfs_ilock(ip);
	error = macfs_iaccess(ip, mode);
	macfs_iunlock(ip);
	return (error);
}

int macfs_getattr(
	struct vnode	*vp,
	register struct vattr *vap,
	struct ucred	*cred)
{
	register struct macfs_node *ip = VTO_MAC(vp);
	int crtime;
	int i;

trace(macfs_debug,("\nDgetattr%x ", ip->o_number));

	vap->va_uid = ip->o_uid;
	vap->va_gid = ip->o_gid;
	vap->va_fsid = ip->o_dev;
	vap->va_nodeid = ip->o_number;
	vap->va_nlink = ip->o_nlink;
	vap->va_blocksize = ip->i_mnt->mrec.vib.AlBlkSiz;

	vap->va_mode = ip->o_mode;
	if(ip->crec.cdrType == CDR_FILE)
	{
	    vap->va_size = ip->crec.U.cdrFilRec.filLgLen;
	    /* access time not available, so modification time */
	    vap->va_atime.tv_sec = TIME_MACTOUX(ip->crec.U.cdrDirRec.dirMdDat);
	    vap->va_atime.tv_usec = 0;
	    vap->va_mtime.tv_sec = TIME_MACTOUX(ip->crec.U.cdrDirRec.dirMdDat);
	    vap->va_mtime.tv_usec = 0;
	    vap->va_ctime.tv_sec = TIME_MACTOUX(ip->crec.U.cdrDirRec.dirCrDat);
	    vap->va_ctime.tv_usec = 0;
	}
	else
	{
	    vap->va_size = ip->crec.U.cdrFilRec.filLgLen;
	    /* access time not available, so modification time */
	    vap->va_atime.tv_sec = TIME_MACTOUX(ip->crec.U.cdrFilRec.filMdDat);
	    vap->va_atime.tv_usec = 0;
	    vap->va_mtime.tv_sec = TIME_MACTOUX(ip->crec.U.cdrFilRec.filMdDat);
	    vap->va_mtime.tv_usec = 0;
	    vap->va_ctime.tv_sec = TIME_MACTOUX(ip->crec.U.cdrFilRec.filCrDat);
	    vap->va_ctime.tv_usec = 0;
	}
	vap->va_rdev = vp->v_rdev;
	vap->va_blocks = (vap->va_size + 1023) >> 10;
	return (0);
}

/*
 * Vnode op for reading/writing
 */
int macfs_rdwr(
	struct vnode	*vp,
	register struct uio *uio,
	enum uio_rw	rw,
	int		ioflag,
	struct ucred	*cred)
{
	register struct macfs_node *ip = VTO_MAC(vp);
	register struct macfs_mount *imp;
	int invec, offset;
	int error = 0;
	uid_t old_VOPuid;	/* xxx I donno whats this for */

trace(macfs_debug,("Drdwr %x %c %x ", ip->o_number, (rw==UIO_READ)?'R':'W', uio->uio_resid));

	if(ip->crec.cdrType != CDR_FILE)
	    return(EINVAL);
	if (rw == UIO_READ && uio->uio_resid == 0)
		return (0);
	if ((ioflag & IO_APPEND) && (rw == UIO_WRITE)) {
	    /*
	     * in append mode start at end of file.
	     */
	    offset = ip->crec.U.cdrFilRec.filLgLen;
	}
	if (((int) uio->uio_offset < 0 ||
	     (int) (uio->uio_offset + uio->uio_resid) < 0))
		return (EINVAL);
/*	if (rw == UIO_READ)
		imark(ip, IACC); */
	if (uio->uio_resid == 0)
		return (0);
	if (rw == UIO_WRITE
	    /*&& type == IFREG*/
	    && uio->uio_offset + uio->uio_resid
	       > u.u_rlimit[RLIMIT_FSIZE].rlim_cur) {
	    psignal(u.u_procp, SIGXFSZ);
	    return (EFBIG);
	}

	old_VOPuid = u.u_VOPuid;
	u.u_VOPuid = cred->cr_uid;

	macfs_ilock(ip);
	/* now have fun with IO: */
	imp = ip->i_mnt;
	offset = uio->uio_offset;
	for(invec = 0; invec < uio->uio_iovcnt; invec++)
	{
	    LONGWORD nbytes = -1;
	    if(rw == UIO_READ)
	    {
		CallMac(mread(&imp->mrec, &ip->crec, offset,
			      uio->uio_iov[invec].iov_len,
			      (BYTE *)uio->uio_iov[invec].iov_base,
			      &nbytes),
			error = EIO,
			("mread %x %x offset %d len %d nbytes %d",
			 &imp->mrec, &ip->crec, offset,
			 uio->uio_iov[invec].iov_len, nbytes));
	    }
	    else
	    {
		CallMac(mwrite(&imp->mrec, &ip->crec, offset,
			       uio->uio_iov[invec].iov_len,
			       (BYTE *)uio->uio_iov[invec].iov_base,
			       &nbytes),
			error = EIO,
			("mwrite %x %x offset %d len %d nbytes %d",
			 &imp->mrec, &ip->crec, offset,
			 uio->uio_iov[invec].iov_len, nbytes));
	    }
	    if(error)
		goto out;
	    offset += nbytes;
	    if(nbytes != uio->uio_iov[invec].iov_len)
		goto out; /* BOOGER: how to report partial success? */
	}
out:
	macfs_iunlock(ip);
	u.u_VOPuid = old_VOPuid;
	return (error);
}

int macfs_ioctl(
	struct vnode	*vp,
	unsigned int	com,
	caddr_t		data,
	int		fflag,
	struct ucred	*cred)
{
trace(macfs_debug,("Dioctl "));
	return (EINVAL);
}

int macfs_select(
	struct vnode	*vp,
	int		which,
	struct ucred	*cred)
{
trace(macfs_debug,("Dselect "));
	return (EINVAL);
}

/* macro previously used to compute ownership abilities */
#undef OWNER
static int OWNER(struct ucred *cred, struct macfs_node *ip)
{
    if(cred->id_uid == ip->o_uid || cred->id_ruid == ip->o_uid)
	return(0);
    return(EACCES);
}

/*
 * Setattr
 */
int macfs_setattr(
	register struct vnode *vp,
	register struct vattr *vap,
	struct ucred *cred)
{
    register struct macfs_node *ip;
    int chtime = 0;
    int error = 0;

trace(macfs_debug,("Dsetattr %x ", ip->o_number));
    /*
     * cannot set these attributes
     */
    if ((vap->va_nlink != -1) || (vap->va_blocksize != -1) ||
	(vap->va_rdev != -1) || (vap->va_blocks != -1) ||
	(vap->va_fsid != -1) || (vap->va_nodeid != -1)) {
	return (EINVAL);
    }

    ip = VTO_MAC(vp);
    macfs_ilock(ip);
    /*
     * Change file access modes.
     */
    if (vap->va_mode != (u_short)-1) {
#if 0
	/* sleazy */
#undef i_uid
#define i_uid o_uid
	if(error = OWNER(cred, ip))
	    goto out;
	ip->o_mode &= IFMT;
	ip->o_mode |= vap->va_mode & ~IFMT;
	if (cred->cr_uid != 0) {
	    if ((ip->o_mode & IFMT) != IFDIR)
		ip->o_mode &= ~ISVTX;
	    if (!groupmember(ip->o_gid))
		ip->o_mode &= ~ISGID;
	}
#if 0
		imark(ip, ICHG);
		if ((vp->v_flag & VTEXT) && ((ip->o_mode & ISVTX) == 0)) {
			if (ip->i_header.pager != MEMORY_OBJECT_NULL)
			        inode_uncache(vp);
		    }
#endif
#endif
	error = ENODEV;
	goto out;
    }
    /*
     * To change file ownership, must be su.
     * To change group ownership, must be su or owner and in target group.
     * This is now enforced in chown1() below.
     */
    /* va_uid and va_gid are short, not unsigned (ie uid_t and gid_t) */
    if ((vap->va_uid != -1) || (vap->va_gid != -1)) {
	if(error = macfs_chown1(ip, vap->va_uid, vap->va_gid))
	    goto out;
    }

    /*
     * Truncate file. Must have write permission and not be a directory.
     */
    if (vap->va_size != -1) {
	if ((ip->o_mode & IFMT) == IFDIR) {
	    error = EISDIR;
	    goto out;
	}
	if(error = macfs_iaccess(ip, IWRITE))
	    goto out;
	if(error = macfs_itrunc(ip, vap->va_size))
	    goto out;
    }
#if 0
#if	MACH_NBC
    /*
     *	Sync out all blocks to prevent delayed writes from
     *	changing the modified time later.  Need to unlock
     *	the inode so the pager can page out the pages... this
     *	is a bit hokey but we'll fix it when we fix
     *	the rest of the filesystem.
     */
    macfs_iunlock(ip);
    (void) mfs_fsync(vp);
    macfs_ilock(ip);
#endif	MACH_NBC
#endif
    /*
     * Change file access or modified times.
     */
    if (vap->va_atime.tv_sec != -1) {
	if(error = OWNER(cred, ip))
	    goto out;
	if(ip->crec.cdrType == CDR_FILE)
	    ip->crec.U.cdrDirRec.dirMdDat
		= TIME_UXTOMAC(vap->va_atime.tv_sec);
	else
	    ip->crec.U.cdrFilRec.filMdDat
		= TIME_UXTOMAC(vap->va_atime.tv_sec);
	chtime++;
    }
    if (vap->va_mtime.tv_sec != -1) {
	if(error = OWNER(cred, ip))
	    goto out;
	if(ip->crec.cdrType == CDR_FILE)
	    ip->crec.U.cdrDirRec.dirMdDat
		= TIME_UXTOMAC(vap->va_mtime.tv_sec);
	else
	    ip->crec.U.cdrFilRec.filMdDat
		= TIME_UXTOMAC(vap->va_mtime.tv_sec);
	chtime++;
    }
    if (chtime) {
	struct timeval time;
	get_time(&time);
#if 0
		ip->i_flag |= IACC|IUPD|ICHG;
#endif
	if(ip->crec.cdrType == CDR_FILE)
	    ip->crec.U.cdrDirRec.dirCrDat
		= TIME_UXTOMAC(time.tv_sec);
	else
	    ip->crec.U.cdrFilRec.filCrDat
		= TIME_UXTOMAC(time.tv_sec);
    }
    macfs_iupdat(ip, 1);	/* XXX should be asyn for perf */
out:
    macfs_iunlock(ip);
    return (error);
}

/*
 * Perform chown operation on inode ip;
 * inode must be locked prior to call.  Right now this can't be done --
 * no per-file access anything.
 */
int macfs_chown1(ip, uid, gid)
	register struct macfs_node *ip;
	register uid_t uid;
	gid_t gid;
{
    return(ENODEV);
}

/* underimplemented */
int macfs_create(
	struct vnode	*dvp,
	char		*nm,
	struct vattr	*vap,
	enum vcexcl	exclusive,
	int		mode,
	struct vnode	**vpp,
	struct ucred	*cred)
{
    return(EINVAL);
}

/* underimplemented */
int macfs_remove(
	struct vnode	*vp,
	char		*nm,
	struct ucred	*cred)
{
    return(EINVAL);
}

/* underimplemented */
int macfs_rename(
	struct vnode	*sdvp,		/* old (source) parent vnode */
	char		*snm,		/* old (source) entry name */
	struct vnode	*tdvp,		/* new (target) parent vnode */
	char		*tnm,		/* new (target) entry name */
	struct ucred	*cred)
{
    return(EINVAL);
}

/* underimplemented */
int macfs_mkdir(
	struct vnode	*dvp,
	char		*nm,
	register struct vattr *vap,
	struct vnode	**vpp,
	struct ucred	*cred)
{
    return(EINVAL);
}

/* underimplemented */
int macfs_rmdir(
	struct vnode	*vp,
	char		*nm,
	struct ucred	*cred)
{
    return(EINVAL);
}

/*
 * Vnode op for readdir
 */
int macfs_readdir(
	struct vnode	*vp,
	register struct uio *uio,
	struct ucred	*cred)
{
	struct direct dirent;
	int macfs_offset;
	int entryoffsetinblock;
	int error = 0;
	int nchild, nchildren;
	CATRECANDNAME *children = NULL;
	int endsearch;
	struct macfs_mount *imp;
	struct macfs_node *ip;
	int nBytesOut = 0;


trace(macfs_debug,("\nDreaddir %x\n", uio->uio_offset));
	ip = VTO_MAC (vp);
	imp = ip->i_mnt;

	macfs_offset = uio->uio_offset;

	/*
	 * Before we start. MAC does not any '.' or '..' entry, so
         * we invent it.
	 */
	if (macfs_offset == 0) {
		dirent.d_name[0] = '.';
		dirent.d_name[1] = 0;
		dirent.d_namlen = 1;
		dirent.d_ino = ip->o_number;
		dirent.d_reclen = DIRSIZ (&dirent);
		if (uio->uio_resid >= dirent.d_reclen)
		{
		    nBytesOut += dirent.d_reclen;
		    if (error = uiomove(&dirent, dirent.d_reclen,
					UIO_READ, uio))
			return error;
		}
		dirent.d_name[0] = '.';
		dirent.d_name[1] = '.';
		dirent.d_name[2] = 0;
		dirent.d_namlen = 2;
		dirent.d_ino = ip->ParentId;
		dirent.d_reclen = DIRSIZ (&dirent);
		if (uio->uio_resid >= dirent.d_reclen)
		{
		    nBytesOut += dirent.d_reclen;
		    if (error = uiomove(&dirent, dirent.d_reclen,
					UIO_READ, uio))
			return error;
		}
	}
	trace(macfs_debug, ("Call mdirentries %x %d",
			    &imp->mrec, ip->crec.U.cdrDirRec.dirDirID));
	nchildren = mdirentries(&imp->mrec,
				  ip->crec.U.cdrDirRec.dirDirID, &children);
	trace(macfs_debug, ("Rtrn mdirentries %x %d = %d",
			    &imp->mrec, ip->crec.U.cdrDirRec.dirDirID,
			    nchildren));
	if(nchildren < 0)
	    return(error = nchildren);
	for(nchild = 0; nchild < nchildren; nchild++)
	{
	    CATALOG_RECORD *cr = &children[nchild].cr;
	    /* Don't need to worry about overlong names */
	    strcpy(dirent.d_name, children[nchild].name);
	    dirent.d_namlen = strlen(dirent.d_name);
	    if(cr->cdrType == CDR_DIR)
		dirent.d_ino = cr->U.cdrDirRec.dirDirID;
	    else if(cr->cdrType == CDR_FILE)
		dirent.d_ino = cr->U.cdrFilRec.filFlNum;
	    else
		dirent.d_ino = -1;
	    if(dirent.d_ino >= 0)
	    {
		dirent.d_reclen = DIRSIZ(&dirent);
		if (uio->uio_resid < dirent.d_reclen)
		    break;
		if(nBytesOut >= macfs_offset)
		{
		    macfs_offset += dirent.d_reclen;
		    if(error = uiomove(&dirent, dirent.d_reclen,
				       UIO_READ, uio))
			goto out;
		}
		nBytesOut += dirent.d_reclen;
	    }
	}
	uio->uio_offset = macfs_offset;

	out:
	free(children);
	return (error);
}

/*
 * Create a symlink
 */
int macfs_symlink(
	struct vnode	*dvp,
	char		*lnm,
	struct vattr	*vap,
	char		*tnm,
	struct ucred	*cred)
{
	return(EINVAL);
}

/*
 * Read a symlink
*/
int macfs_readlink(
	struct vnode	*vp,
	struct uio	*uiop,
	struct ucred	*cred)
{
	return(EINVAL);
}

/*
 * Just call the block device strategy routine.
 */
int macfs_strategy(
	register struct buf *bp)
{
	register struct macfs_node *ip;
	struct vnode *vp = ITOV(bp->b_vp);
	int error;

trace(macfs_debug,("Dstrategy %x ", bp->b_blkno));

	ip = VTO_MAC(vp);
	vp = ip->i_mnt->im_devvp;
	bp->b_dev = vp->v_rdev;
	(*_VOP_(vp)->vn_strategy)(bp);
	return (0);
}

/* I don't think we need
   macfs_nfsbmap, macfs_brelse, macfs_pagein, macfs_pageout */

int macfs_nlinks(
	struct vnode	*vp,
	int		*l,
	struct ucred	*cred)
{
trace(macfs_debug,("Dnlinks "));
	*l = VTO_MAC(vp)->o_nlink;
	return (0);
}

/*
 * Update disk status of inode
 */
int macfs_vupdate(
	struct vnode	*vp,
	int		flag)
{
	macfs_ilock(VTO_MAC(vp));
	macfs_iupdat(VTO_MAC(vp),flag);
	macfs_iunlock(VTO_MAC(vp));
	return 0;
}

/*
 * See original ufs_lookup for structure.
 */
int macfs_lookup(
	struct vnode	*vdp,
	char		*nm,
	struct vnode	**vpp,
	struct ucred	*cred)
{
	struct vnode *vp;
	register int error;

trace(macfs_debug,("\nDlookup %s ", nm));
	error = macfs_dirlook(VTO_MAC(vdp), nm, &vp, cred);
	if (error == 0) {
		*vpp = vp;
		macfs_iunlock(VTO_MAC(vp));
#if 0
		/*
		 * If vnode is a device return special vnode instead
		 */
		if (ISVDEV(((*vpp)->v_mode&VFMT))) {
			struct vnode *newvp;

			newvp = specvp(*vpp, (*vpp)->v_rdev);
			VN_RELE(*vpp);
			*vpp = newvp;
		}
#endif
	}
	return (error);
}

int macfs_dirlook(
	register struct macfs_node *dp,	/* the directory we are searching */
	char		*nm,		/* the name we are looking for */
	struct vnode	**vpp,		/* the vnode we return */
	struct ucred	*cred)		/* credentials to use */
{
    int error;
    
trace(macfs_debug,("Ddirlook %x\n", dp->o_number));
    if (dp->crec.cdrType != CDR_DIR)
	return (ENOTDIR);
    /* don't need to check read permissions?? */
    if(strcmp(nm, "."))
    {
	macfs_ilock(dp);
	VN_HOLD(MAC_TOV(dp));
	*vpp = MAC_TOV(dp);
	return(0);
    }
    else if(strcmp(nm, ".."))
    {
	*vpp = MAC_TOV(macfs_finddir(dp->i_mnt, dp->ParentId));
	return(*vpp? 0 : EINVAL);
    }
    else
    {
	struct macfs_node *mp = macfs_getnode(dp->i_mnt, dp->o_number, nm);
	if(mp == NULL) /* lookup failed */
	{
	    CATALOG_RECORD crec;
	    /* Btree lock() will protect this */
	    CallMac(mopen(&dp->i_mnt->mrec, dp->crec.U.cdrDirRec.dirDirID,
			  nm, &crec),
		    return(ENOENT),
		    ("mopen %x %d %s", &dp->i_mnt->mrec,
		     dp->crec.U.cdrDirRec.dirDirID, nm));
	    if(!(mp = crec_to_macfsnode(dp->i_mnt, &crec,
					dp->crec.U.cdrDirRec.dirDirID,
					nm)))
		return(ENOENT);
	    *vpp = MAC_TOV(mp);
	    return(0);
	}
    }
    return(0);
}

struct macfs_node *macfs_finddir(struct macfs_mount *i_mnt, LONGWORD did)
{
    struct macfs_node *mp;
    CATALOG_RECORD crec;
    
    /* Cannot use cache lookup in this routine -- have
       no parent vnode */
    if(did == ROOTID)
    {
	CallMac(mopenroot(&i_mnt->mrec, &crec, NULL),
		return(NULL),
		("mopenroot %x ", &i_mnt->mrec));
	if(!(mp = crec_to_macfsnode(i_mnt, &crec, ROOTID, SMALLESTFN)))
	    return(NULL);
	return(mp);
    }
    else
    {
	int error;
	LONGWORD ParentId;
	char *Name;

	CallMac(mopen(&i_mnt->mrec, did, SMALLESTFN, &crec),
		return(NULL),
		("mopen %x %d SMALLESTFN",
		 &i_mnt->mrec, did));
	if(crec.cdrType != CDR_DTHREAD)
	{
	    trace(macfs_debug, ("thread record on dir %d: cdrType = %d",
				crec.cdrType));
	    return(NULL);
	}
	ParentId = crec.U.cdrThdRec.thdParID;
	Name = crec.U.cdrThdRec.thdCName;
	    
	CallMac(mopen(&i_mnt->mrec, ParentId, Name, &crec),
		return(NULL),
		("mopen %x %d %s", &i_mnt->mrec, ParentId, Name));
	if(!(mp = crec_to_macfsnode(i_mnt, &crec, ParentId, Name)))
	    return(NULL);
	/* sanity check */
	if(mp->o_number != did)
	{
	    printf("macfs_finddir: found directory_id %lu instead of %lu\n",
		   (unsigned long)mp->o_number,
		   (unsigned long)did);
	    return(NULL);
	}
	return(mp);
    }
    return(NULL); /* UNREACHABLE */
}

zone_t macfs_zone;

void macfs_init(void)
{
	register int i;
	register union macfs_ihead *ih = macfs_ihead;

	for (i = INOHSZ; --i >= 0; ih++) {
		ih->ih_head[0] = ih;
		ih->ih_head[1] = ih;
	}
	macfs_zone = zinit(sizeof(struct macfs_node), 1024*1024,
				0, FALSE, "macfs zone");
/*	macfs_fid_zone = zinit(sizeof(struct ifid), 1024*1024,
				vm_page_size, FALSE, "macfs fid");*/
}

/* if returns 1, call this again -- there was a race failure.
   if returns 0, *mp NULL --> not found, else *mp is the desired
   answer, and is locked and held */
int macfs_dev_and_num_lookup(dev_t dev, int o_number, struct macfs_node **mp)
{
    union macfs_ihead *ih;
    struct macfs_node *ip;
    for(ip = ih->ih_chain[0];
	ip != (struct macfs_node *)ih;
	ip = (struct macfs_node *)ip->i_forw)
	if(ip->o_number == o_number && ip->o_dev == dev)
	{
	    macfs_ilock(ip);	
	    /* perhaps we lost the waiting race? */
	    if(ip->o_number != o_number || ip->o_dev != dev)
	    {
		macfs_iunlock(ip);
		return(1); /* race failure */
	    }
	    *mp = ip;
	    VN_HOLD(MAC_TOV(ip));
	    return(0);
	}
    *mp = (struct macfs_node *)NULL;
    return(0);
}

/* non_null return value is always locked and held */
struct macfs_node *crec_to_macfsnode(struct macfs_mount *i_mnt,
				     CATALOG_RECORD *crec,
				     int ParentId,
				     char *name)
{
    union macfs_ihead *ih;
    struct macfs_node *ip;
    int o_number;
    if(crec->cdrType == CDR_FILE)
	o_number = crec->U.cdrFilRec.filFlNum;
    else if(crec->cdrType == CDR_DIR)
	o_number = crec->U.cdrDirRec.dirDirID;
    else
	return(NULL);
    while(macfs_dev_and_num_lookup(i_mnt->im_dev, o_number, &ip))
	;
    if(ip)
	return(ip);
    /* so how do we provide mutual exclusion for this code?
       Dosfs doesn't seem to consider it necessary. */
    ih = &macfs_ihead[INOHASH(i_mnt->im_dev, o_number)];
    ZALLOC(macfs_zone, ip, struct macfs_node *);
    bzero(ip, sizeof(*ip));
    MAC_TOV(ip)->v_type = ITYPE_MACFS;
    ip->o_dev = i_mnt->im_dev;
    ip->o_number = o_number;
    VN_HOLD(MAC_TOV(ip));
    macfs_ilock(ip);
    insque(ip, ih);

    /* now all that special macfs initialization */
    memcpy(&ip->crec, &crec, sizeof(crec));
    ip->i_mnt = i_mnt;

    /* now init the vnode */
    VN_INIT(MAC_TOV(ip), i_mnt->vfsp, 0, i_mnt->im_dev);
    if(ip->crec.cdrType == CDR_DIR)
    {
	ip->o_nlink = 2;
	MAC_TOV(ip)->v_mode = VDIR;
    }
    else if(ip->crec.cdrType == CDR_FILE)
    {
	ip->o_nlink = 1;
	MAC_TOV(ip)->v_mode = VREG;
    }
    else
    {
	ip->o_nlink = 0;
	macfs_iunlock(ip);
	VN_RELE(ip);
	printf("Bad crec passed into crec_to_macfsnode\n");
	return(NULL); /* serious error */
    }
    ip->o_uid = i_mnt->im_uid;
    ip->o_gid = i_mnt->im_gid;
    ip->o_mode = i_mnt->im_mode | MAC_TOV(ip)->v_mode;
    
    if(o_number == ROOTID)
	MAC_TOV(ip)->v_flag |= VROOT;
    if(i_mnt->im_devvp)
	VN_HOLD(i_mnt->im_devvp);
    return(ip);
}

/* will use the Directory Name Lookup Cache (dnlc) in the future.
   Return val is always locked and held. */
struct macfs_node *macfs_getnode(struct macfs_mount *i_mnt,
				 LONGWORD ParentId,
				 char *name)
{
    return(NULL);
}

/*
 * Last reference to an inode, write the inode out and if necessary,
 * truncate and deallocate the file.
 */
int macfs_inactive(
	struct vnode	*vp,
	struct ucred	*cred)
{
    register struct macfs_node *ip = VTO_MAC(vp);
    int mode, error = 0;
trace(macfs_debug,("Dinactive %x %x\n", ip->o_number, ip->o_nlink));
    macfs_ilock(ip);
    macfs_iupdat(ip, 1);
    return macfs_zfree(ip);
}

/* Put an inode to rest.  Inode must be locked */
int macfs_zfree(struct macfs_node *ip)
{
    trace(macfs_debug, ("Dzfree %x ", ip->o_number));
    remque(ip);
    ip->i_forw = (struct vnode *)ip;
    ip->i_back = (struct vnode *)ip;
/*    dnlc_purge_vp(MAC_TOV(ip));*/
    if (ip->i_mnt->im_devvp)
	VN_RELE(ip->i_mnt->im_devvp);

#if 0
    /*we should do something like this when we support deleting files*/
    if (ip->o_nlink <= 0)
	if (ip->i_size == 0) dosfs_igone(ip);
#endif

    /*
     * wakeup anyone that raced us to
     * this inode in crec_to_macfsnode, and lost
     */
    macfs_iunlock(ip);
    /* put it back on freelist or zone */
    ZFREE(macfs_zone, ip);

    return (0);
}

int macfs_fid(struct vnode *vp, struct fid **fidpp)
{
    return(ENODEV);
}

int macfs_freefid(struct vnode *vp, struct fid *fidp)
{
    return(ENODEV);
}

int macfs_fhtovp(struct vfs *vfsp, struct vnode	**vpp, struct fid *fhp)
{
    return(ENODEV);
}

int macfs_itrunc(struct macfs_node *ip, int size)
{
    return(ENODEV);
}

/*
 * Access check
 */
int macfs_iaccess(
	struct macfs_node *ip,
	int		  mode)
{
	int m;
	gid_t *gp;

	if(mode & IEXEC)
	    return(EACCES); /* we can't support execution */
	if (u.u_uid == 0) /* superuser */
	    return(0);
	m = mode;
	if (u.u_uid != ip->o_uid) {
	    m >>= 3;
	    if (u.u_gid == ip->o_gid)
		goto found;
	    gp = u.u_groups;
	    for (; gp < &u.u_groups[NGROUPS] && *gp != NOGROUP;
		 gp++)
		if (ip->o_gid == *gp)
		    goto found;
	    m >>= 3;
	}
found:
	if ((ip->o_mode & m) == m)
	    return (0);
	return (u.u_error = EACCES);
}

void macfs_iupdat(
	struct macfs_node	*ip,
	int			waitfor)
{
    /* if anyone can write */
    if(ip->i_mnt->im_mode & (VWRITE | (VWRITE>>3) | (VWRITE>>6)))
    {
	CallMac(mclose(&ip->i_mnt->mrec, ip->ParentId, ip->name, &ip->crec),
		;,
		("mclose %x %d %s %x", &ip->i_mnt->mrec, ip,
		 ip->ParentId, ip->name));
    }
    else
	trace(macfs_debug, ("Ignoring update request on %x %x",
			    &ip->i_mnt->mrec, ip));
}

void macfs_irelease(
	struct macfs_node	*ip)
{
	if (ip->i_flag & ILOCKED)
		panic("macfs_irelease");
	VN_RELE(MAC_TOV(ip));
}

/*
 * Lock an inode. If its already locked, set the WANT bit and sleep.
 */
void macfs_ilock(ip)
	register struct macfs_node *ip;
{
trace(macfs_debug,("L %x ", ip->o_number));

    while (ip->i_flag & ILOCKED) {
	ip->i_flag |= IWANT;
	(void) sleep((caddr_t)ip, PINOD);
    }
    ip->i_flag |= ILOCKED;

}

/*
 * Unlock an inode.  If WANT bit is on, wakeup.
 */
void macfs_iunlock(ip)
	register struct macfs_node *ip;
{
trace(macfs_debug,("U %x ", ip->o_number));

    if ((ip->i_flag & ILOCKED) == 0)
	printf("macfs_iunlock: unlocked inode", MAC_TOV(ip));
    ip->i_flag &= ~ILOCKED;
    if (ip->i_flag&IWANT) {
	ip->i_flag &= ~IWANT;
	wakeup((caddr_t)ip);
    }
}

int macfs_iflush(dev_t dev)
{
	register struct macfs_node *ip, *iq;
	register open = 0;
	union macfs_ihead *ih;

	for (ih = macfs_ihead; ih < &macfs_ihead[INOHSZ]; ih++) {
	  iq = 0;
retry:
	  for(ip = ih->ih_chain[0];
	      ip != (struct macfs_node *)ih;
	      ip = (struct macfs_node *)ip->i_forw) {

		if (ip->o_dev == dev) {
			/*
			 * Inode in active use. We tried to flush
			 * it and failed. Give up.
			 */
			if (ip == iq)
				return (-1);
			/*
			 * See if the inode pager can rid of it
			 */
			if (inode_uncache_try(ip))
				/* program is active */
				return (-1);
			/*
			 * Inode pager either knows nothing about it,
			 * or was able to uncache it.  To avoid locking
			 * other inodes just make a note and rescan
			 * the chain; the inode will not be on any
			 * chain if we flushed it above.
			 */
			iq = ip;
			goto retry;
		}
		else if ((ip->o_mode&IFMT)==IFBLK &&
		    /* xxxx is this right xxxx */
		    ip->i_vnode.v_rdev == dev)
			open++;
	  }
	}
	return (open);
}

unsigned long now()
{
    struct timeval tv;
    get_time(&tv);
    return(tv.tv_sec + 0x7C25783F);
}

int initreadwrite(struct vnode *device, int partnum, MOUNT_RECORD *mr)
{
  int rc;

  if (device==NULL)
    return INITREADWRITEFAILED;

  mr->sr.device=device;
  mr->sr.validflag=VALIDVAL;
  mr->sr.maxwritesbeforemrflush=MAXWRITESBEFOREMRFLUSH;
  mr->sr.currentnumwrites=0;


  if ((rc=getpartinfo(mr,partnum))) {
    mr->sr.validflag=0;
    return rc;
  }

  return 0;
}

int readabsolutesector(MOUNT_RECORD *mr, SECTNUMTYPE sect, char *buf)
{
  struct staterec *what;
  struct buf *bp;

  what=&(mr->sr);

  if (what->validflag!=VALIDVAL)
    return INVALIDSTATEREC;

  bp=NULL;
  bp=bread(what->device, sect, what->sectsize);
  if (bp->b_error)
    return READSECTFAILED;
  bcopy(bp->b_un.b_addr,buf,what->sectsize);
  brelse(bp);
  
  return 0;
}

int readsects(MOUNT_RECORD *mr, SECTNUMTYPE sect, int num, char *buf)
{
  struct staterec *what;
  SECTNUMTYPE cursect;
  struct buf *bp;
  BYTE *bufptr;

  what=&(mr->sr);

  sect+=what->sectoffsettopartition;

  if (what->validflag!=VALIDVAL)
    return INVALIDSTATEREC;

  bufptr=buf;

  for (cursect=sect;cursect<(sect+num);cursect++, bufptr+=what->sectsize) {
    bp=NULL;
    bp=bread(what->device, cursect, what->sectsize);
    if (bp->b_error)
      return READSECTFAILED;
    bcopy(bp->b_un.b_addr,bufptr,what->sectsize);
    brelse(bp);
  }
  
  return num;

}

int writesects(MOUNT_RECORD *mr, SECTNUMTYPE sect, int num, char *buf)
{
  struct staterec *what;
  int rc;
  SECTNUMTYPE cursect;
  struct buf *bp;
  BYTE *bufptr;

  what=&(mr->sr);

  sect+=what->sectoffsettopartition;


  if (what->validflag!=VALIDVAL)
    return INVALIDSTATEREC;

  
  bufptr=buf;

  for (cursect=sect;cursect<(sect+num);cursect++, bufptr+=what->sectsize) {
    bp=NULL;
    bp=getblk(what->device, cursect, what->sectsize);
    if (bp->b_error)
      return WRITESECTFAILED;
    bcopy(bufptr, bp->b_un.b_addr, what->sectsize);
    bdwrite(bp);
  }

#ifdef USER
  if (what->currentnumwrites>=what->maxwritesbeforemrflush) {
    if ((rc=syncmountrecord(mr))) {
      return rc;
    } else {
      what->currentnumwrites=0;
    }
  } else {
    what->currentnumwrites+=num;
  }
#endif
  
  return num;

}

int writesectsunbuf(MOUNT_RECORD *mr, SECTNUMTYPE sect, int num, char *buf)
{
  struct staterec *what;
  SECTNUMTYPE cursect;
  BYTE *bufptr;
  struct buf *bp;


  what=&(mr->sr);

  sect+=what->sectoffsettopartition;


  if (what->validflag!=VALIDVAL)
    return INVALIDSTATEREC;
  
  bufptr=buf;

  for (cursect=sect;cursect<(sect+num);cursect++, bufptr+=what->sectsize) {
    bp=NULL;
    bp=getblk(what->device, cursect, what->sectsize);
    if (bp->b_error)
      return WRITESECTFAILED;
    bcopy(bufptr, bp->b_un.b_addr, what->sectsize);
    bwrite(bp);
  }


  return num;
}
